import re
import time
from collections import namedtuple

import psutil

device_iostats = namedtuple('device_iostats', 'read_time write_time busy_time')

def make_device_iostats(old_stats, new_stats, timediff):
    read_time = (new_stats.read_time - old_stats.read_time) * 100 / timediff
    write_time = (new_stats.write_time - old_stats.write_time) * 100 / timediff
    try:
        busy_time = (new_stats.busy_time - old_stats.busy_time) * 100 / timediff
    except:
        busy_time = read_time + write_time
    return device_iostats(read_time, write_time, busy_time)

def is_valid_name(name: str) -> bool:
    return re.fullmatch(r"nvme\dn\d|sd\w|disk\d", name)

class IOStat:
    def __init__(self):
        self._stats = {}
        self._time = None

        self.keys = []
        self.stats = {}


    def setup(self):
        self._update()
        self.keys = self._stats.keys()

    def update(self):
        old_stats, old_time = self._stats, self._time
        self._update()
        new_stats, new_time = self._stats, self._time
        timediff_ms = (new_time - old_time) * 1000

        keys = set(old_stats.keys()) & set(new_stats.keys())
        self.stats = {
            key: make_device_iostats(old_stats[key], new_stats[key], timediff_ms) for key in keys
        }

    def _update(self):
        self._stats = {k: v for k, v in psutil.disk_io_counters(True).items() if is_valid_name(k)}
        self._time = time.time()



import numpy as np

from ..system_stat import MemoryStat
from . import common
from .animated import AnimatedAxes


class MemoryPlotter(AnimatedAxes):
    def __init__(self, mem=None):
        self.mem = mem or MemoryStat()

    def setup(self, axes, x):
        self.mem.setup()
        self.y_mem = common.none_array(x.size)
        self.y_swap = common.none_array(x.size)
        self.line_mem = axes.plot(x, self.y_mem, label='Memory')[0]
        self.line_swap = axes.plot(x, self.y_swap, label='Swap')[0]
        self.lines = [self.line_mem, self.line_swap]

        axes.set_title('Memory')
        axes.set_ylabel('% Memory')
        axes.set_xlim(x.min(), x.max())
        axes.set_ylim(0, 100)
        axes.tick_params('x', bottom=False, labelbottom=False)
        axes.grid(True, axis='y')
        axes.legend()

        return self.lines

    def update(self):
        self.mem.update()

        common.additem_cyclic_inplace(self.y_mem, self.mem.mem.percent)
        common.additem_cyclic_inplace(self.y_swap, self.mem.swap.percent)

        self.line_mem.set_ydata(self.y_mem)
        self.line_swap.set_ydata(self.y_swap)

        return self.lines



import numpy as np

from ..system_stat import IOStat
from . import common
from .animated import AnimatedAxes


class IOPlotter(AnimatedAxes):
    def __init__(self, keys=None, io=None):
        self.io = io or IOStat()
        self.keys = keys

    def setup(self, axes, x):
        self.io.setup()
        self.keys = self.keys or self.io.keys
        self.ys = [common.none_array(x.size) for _ in self.keys]
        self.lines = [axes.plot(x, y, label=key)[0] for key, y in zip(self.keys,self.ys)]

        axes.set_title('IO')
        axes.set_ylabel('% IO time')
        axes.set_xlim(x.min(), x.max())
        axes.set_ylim(0, 100)
        axes.tick_params('x', bottom=False, labelbottom=False)
        axes.grid(True, axis='y')
        axes.legend()

        return self.lines

    def update(self):
        self.io.update()

        for key, y, line in zip(self.keys, self.ys, self.lines):
            common.additem_cyclic_inplace(y, self.io.stats[key].busy_time)

            line.set_ydata(y)

        return self.lines



import re
import subprocess

import gpustat


class GPUStat:
    def __init__(self):
        self.gpus = []
        self.utilizations = []

    def setup(self):
        self.gpus = [gpu.name for gpu in gpustat.new_query()]

    def update(self):
        self.utilizations = [gpu.utilization for gpu in gpustat.new_query()]



import psutil


class CPUStat:
    def __init__(self):
        self.user = None
        self.nice = None
        self.system = None
        self.idle = None
        self.utilization = None


    def setup(self):
        self._update()

    def update(self):
        self._update()

    def _update(self):
        stat = psutil.cpu_times_percent()
        self.user = stat.user
        self.nice = stat.nice
        self.system = stat.system
        self.idle = stat.idle
        self.utilization = 100 - stat.idle



import numpy as np

from ..system_stat import GPUStat
from . import common
from .animated import AnimatedAxes


class GPUPlotter(AnimatedAxes):
    def __init__(self, gpu=None):
        self.gpu = gpu or GPUStat()

    def setup(self, axes, x):
        self.gpu.setup()

        self.ys = [common.none_array(x.size) for _ in self.gpu.gpus]

        axes.set_title('GPU')
        axes.set_ylabel('% GPU')
        axes.set_xlim(x.min(), x.max())
        axes.set_ylim(0, 100)
        axes.tick_params('x', bottom=False, labelbottom=False)
        axes.grid(True, axis='y')

        self.lines = [
            axes.plot(x, y, label=gpu)[0] for y, gpu in zip(self.ys, self.gpu.gpus)
        ]
        axes.legend()
        return self.lines

    def update(self):
        self.gpu.update()

        for y, utilization, line in zip(self.ys, self.gpu.utilizations, self.lines):
            common.additem_cyclic_inplace(y, utilization)
            line.set_ydata(y)
        return self.lines



import re
import time
from collections import namedtuple

import psutil

netstat = namedtuple('netstat', 'rx tx')

def make_netstat(old_stats, new_stats, timediff):
    rx = (new_stats.bytes_recv - old_stats.bytes_recv) / timediff
    tx = (new_stats.bytes_sent - old_stats.bytes_sent) / timediff
    return netstat(rx, tx)

def is_valid_device(name, stat):
    if not re.fullmatch(r"en\d+|enp\d+s\d|wlp\d+s\d+f\d+", name):
        return False
    if stat.bytes_recv == 0 and stat.bytes_sent == 0:
        return False
    return True

class NetworkStat:
    def __init__(self):
        self._stats = None
        self._time = None

        self.keys = None
        self.stats = None

    def setup(self):
        self._update()
        self.keys = sorted(self._stats.keys())

    def update(self):
        old_stats, old_time = self._stats, self._time
        self._update()
        new_stats, new_time = self._stats, self._time
        timediff_s = new_time - old_time

        keys = set(old_stats.keys()) & set(new_stats.keys())
        self.stats = {
            key: make_netstat(old_stats[key], new_stats[key], timediff_s) for key in keys
        }

    def _update(self):
        self._stats = {
            k: v for k, v in psutil.net_io_counters(pernic=True).items() if is_valid_device(k, v)
        }
        self._time = time.time()



import math

import matplotlib.pyplot as plt
import numpy as np

from ..system_stat import NetworkStat
from . import common
from .animated import AnimatedAxes


class NetworkPlotter(AnimatedAxes):
    def __init__(self, keys=None, net=None):
        self.net = net or NetworkStat()
        self.keys = keys
        self.ylim = 1024

    def setup(self, axes, x):
        self.axes = axes
        self.net.setup()
        self.keys = self.keys or self.net.keys
        self.labels = []
        for key in self.keys:
            self.labels += [f'{key} RX', f'{key} TX']
        self.ys = [common.none_array(x.size) for _ in self.labels]
        self.lines = [axes.plot(x, y, label=label)[0] for label, y in zip(self.labels, self.ys)]

        axes.set_title('Network')
        axes.set_ylabel('Network IO')
        axes.set_xlim(x.min(), x.max())
        axes.set_ylim(0, self.ylim)
        axes.tick_params('x', bottom=False, labelbottom=False)
        axes.grid(True, axis='y')
        axes.legend()
        axes.yaxis.set_major_locator(plt.LinearLocator(5))
        axes.yaxis.set_major_formatter(plt.FuncFormatter(self._format_ylabel))

        return self.lines

    def update(self):
        self.net.update()

        for label, y, line in zip(self.labels, self.ys, self.lines):
            key, direction = label.split()
            v = self.net.stats[key].rx if direction == 'RX' else self.net.stats[key].tx
            common.additem_cyclic_inplace(y, v)

            line.set_ydata(y)

        self._update_ylim()
        return self.lines

    def _update_ylim(self):
        ymax = 1024
        for y in self.ys:
            if y[-1] != None:
                ymax = max(ymax, y[y!=None].max())
        ylim = 2 ** math.ceil(math.log2(ymax))
        if self.ylim != ylim:
            self.ylim = ylim
            self.axes.set_ylim(0, self.ylim)
            self.axes.figure.canvas.resize_event()

    def _format_ylabel(self, value, tick_number):
        suffixes = ['B', 'KB', 'MB', 'GB', 'TB']
        suffix_index = 0
        while value >= 1024 and suffix_index < len(suffixes):
            value = value / 1024
            suffix_index += 1

        return f'{value:.0f} {suffixes[suffix_index]}'
